;;Urban Warfare
;;
;;Table of contents:
;;SECTION 1: Declarations + controls
;;SECTION 1a: Stray class declarations
;;SECTION 2: Geometry
;;SECTION 3: Display
;;SECTION 4: Mapobjects
;;SECTION 5: City generator
;;SECTION 6: Monsters
;;SECTION 7: Actions
;;SECTION 7a: Combat
;;SECTION 8: Events
;;SECTION 9: Monster loop processing
;;SECTION 10: Main sequence

(in-package :cl-user)

(defpackage :uwar
  (:nicknames :r3)
  (:use :common-lisp :curses)
  (:shadow :time :speed :position))

(in-package :uwar)

;;SECTION 1: Declarations + controls

;;Constants
(defparameter *dx* 1) (defparameter *dy* 1)
(defparameter *width* 60) (defparameter *height* 20)
(defparameter *controls* (make-hash-table :test #'equal))
(defparameter *colors* '(:cblue :cdgreen :ccyan :cred :cpurple :cbrown :cgray
                         :cdark :clblue :cgreen :csky :crose :cpink :cyellow
                         :cwhite))
(defparameter *debug-mode* nil)
(defparameter *epsilon* 1.0e-5)

(defparameter *max-moving-distance* 0.03)
(defparameter *moving-distance* 0.02)
(defparameter *fov-angle* #C(1.0 2.0))
(defparameter *fov-angle-phase* (phase *fov-angle*))


(defparameter *enemy-count* 100)
(defparameter *sight-range* 0.2)
(defparameter *monster-range* 0.4)
(defparameter *intuition-range* 0.001) ;)
(defparameter *hearing-range* 0.06)
(defparameter *hearing-chance* 0.2)
(defparameter *shouting-range* 0.2)
(defparameter *stab-range* 0.03)

(defparameter *maxzoom* 22.0)
(defparameter *minzoom* 0.10)

;;Meta data

(defparameter *player* nil)
(defparameter *allmessages* nil)
(defparameter *kill-count* 0)
(defparameter *turn-count* 0)
(defparameter *zoom-level* 0)

;;Map objects
(defvar *mapobjects* nil)
(defvar *map-sector-hash* nil)
(defvar *monsters* nil)
(defvar *monster-sector-hash* nil)
(defvar *map-minx*)
(defvar *map-miny*)
(defvar *map-maxx*)
(defvar *map-maxy*)

;;Temporary data
(defparameter *curmonster* nil)
(defparameter *quitflag* nil)
(defparameter *messages* nil)
(defparameter *zero-time-action* nil)
(defparameter *cancelled* nil)
(defparameter *curwalls* nil)
(defparameter *curfov* nil)

;;Display

(defparameter *curscreen* (make-array (list *width* *height*)))
(defparameter *tagscreen* (make-array (list *width* *height*)))
(defparameter *monscreen* (make-array (list *width* *height*)))
(defparameter *fovmap*  (make-array (list *width* *height*)
                                    :element-type 'bit))
(defparameter *screen-base* nil)
(defparameter *show-fov* nil)
(defparameter *color-fix* nil)

(defun control-bind (action &rest keys)
  (dolist (i keys)
    (setf (gethash i *controls*) action)))

(defun init-controls ()
  (setf *controls* (make-hash-table :test #'equal))

  ;;Cheat controls
  (when *debug-mode*
    (control-bind (make-instance 'restore-hp) 82 138 )
    (control-bind (make-instance 'show-repl-console) 126 240))

  (if (probe-file "controls.cfg")
    (with-open-file (i "controls.cfg")
      (let* ((*read-eval* nil)
             (*package* (find-package :uwar))
             (remapping (read i)))
        (loop for remap in remapping
             do (apply #'control-bind
                       (if (keywordp (car remap))
                           (car remap)
                           (apply #'make-instance (car remap)))
                       (cdr remap))))
      nil)
    :error))

(defun ubergetch ()
  "Workaround for insanely stupid Ncurses keypad handling"
  (loop with keychord
       for x = (getch)
       do (push x keychord)
       while (eql (gethash (reverse keychord) *controls*) :prefix)
       finally (return (if (= (length keychord) 1) 
                           (car keychord) (nreverse keychord)))))

;;SECTION 1a: Stray class declarations

(defclass event () 
  ((interesting :initform t :initarg :interesting :accessor interesting)
   (pt :initarg :pt :accessor pt)))

(defclass e-step (event) 
  ((interesting :initform nil)
   (monster :initarg :monster :accessor monster)))

(defclass e-wait (event) 
  ((interesting :initform nil)
   (monster :initarg :monster :accessor monster)))

(defclass e-combat (event)
  ((attacker :initarg :attacker :accessor attacker)
   (defender :initarg :defender :accessor defender)))

(defclass e-shout (event)
  ((monster :initarg :monster :accessor monster)
   (pt :initarg :pt :accessor pt)))

;;SECTION 2: Geometry

;;The point is a complex number. x is real part, y is imaginary part.

(defmethod x ((point number))
  (realpart point))

(defmethod y ((point number))
  (imagpart point))

(proclaim '(inline scalar-product anti-product norm))

(defun scalar-product (v1 v2)
  (+ (* (x v1) (x v2)) (* (y v1) (y v2))))

(defun anti-product (v1 v2)
  (- (* (x v1) (y v2)) (* (y v1) (x v2))))

(defun norm (v)
  (+ (* (x v) (x v)) (* (y v) (y v))))

(defclass line ()
  ((start :initarg :start :accessor start)
   (end :initarg :end :accessor end)))

(defgeneric orient (curve point)
  (:documentation "Returns true if point is inside (to the right
  of) the curve."))

;; (defun generic-orient (line point)
;;   (let ((x (x point)) (y (y point))
;;         (x1 (x (start line))) (x2 (x (end line)))
;;         (y1 (y (start line))) (y2 (y (end line))))
;;     (let ((scalar (- (* (- y2 y1) (- x x1)) (* (- x2 x1) (- y y1)))))
;;       (> scalar 0))))

(proclaim '(inline fast-orient))

(defun fast-orient (start end point)
  (declare (optimize (cl:speed 3) (safety 0)))
  (declare (number start end point))
  (> (anti-product (- point start) (- end start)) 0))

(defmethod orient ((line line) point)
  (fast-orient (start line) (end line) point))

(defmethod print-object ((line line) s)
  (print-unreadable-object (line s :type t :identity nil)
    (format s "~a ~a" (start line) (end line))))


(defun make-line (start end)
  (make-instance 'line :start start :end end))

(defgeneric distance (obj1 obj2)
  (:documentation "Distance between various objects")
  (:method (point1 point2)
    "Calculate distance between two points by default"
    (sqrt (+ (* (x point1) (x point2)) (* (y point1) (y point2))))))

(defmethod distance ((p1 number) (p2 number))
  "More efficient version for complex points"
  (abs (- p1 p2)))

(defgeneric intersect (c1 c2)
  (:documentation "Returns a list of intersection points of two curves"))

(defgeneric intersect-p (c1 c2)
  (:documentation "Returns whether two curves intersect"))

(defun generic-intersect-p (line1 line2)
  (let ((os1 (orient line2 (start line1))) (os2 (orient line1 (start line2)))
        (oe1 (orient line2 (end line1))) (oe2 (orient line1 (end line2))))
    (not (or (eql os1 oe1) (eql os2 oe2)))))

(defun generic-intersect (line1 line2)
  "Intersects two lines precisely and other curves by linear
interpolation"
  (let* ((s1 (start line1)) (s2 (start line2))
         (e1 (end line1)) (e2 (end line2)))
    (when (generic-intersect-p line1 line2)
      (let* ((v1 (- e1 s1)) (v2 (- s2 e2)) (z (- s2 s1))
             (v11 (norm v1)) (v22 (norm v2))
             (v12 (scalar-product v1 v2))
             (det (- (* v11 v22) (* v12 v12))))
        ;;(format t "~a ~%" det)
        (unless (< (sqrt (abs det)) *epsilon*)
          (list (+ s1 (* v1 (/ (- (* (scalar-product z v1) v22)
                                  (* (scalar-product z v2) v12))
                               det)))))))))

(defmethod intersect-p ((line1 line) (line2 line))
  (generic-intersect-p line1 line2))
  
(defmethod intersect ((line1 line) (line2 line))
  (generic-intersect line1 line2))

(defun project (v1 v2)
  "Project v1 on v2"
  (* v2 (/ (scalar-product v1 v2) (norm v2))))

(defmethod distance ((point number) (line line))
  (let ((a (- point (start line)))
        (b (- (end line) (start line))))
    (if (<= 0.0 (scalar-product a b) (norm b))
        (abs (- a (project a b)))
        (min (distance point (start line))
             (distance point (end line))))))

(defmethod distance ((line line) (point number))
  (distance point line))

(defclass arc ()
  ((center :initarg :center :accessor center)
   (radius :initarg :radius :accessor radius)
   (start-angle :initarg :start-angle :accessor start-angle)
   (end-angle :initarg :end-angle :accessor end-angle)))

(defmethod start ((arc arc))
  (+ (center arc) (* (radius arc) (signum (start-angle arc)))))

(defmethod end ((arc arc))
  (+ (center arc) (* (radius arc) (signum (end-angle arc)))))

(defmethod orient ((arc arc) point)
  (if (< (anti-product (start-angle arc) (end-angle arc)) 0)
      (> (radius arc) (distance point (center arc)))
      (< (radius arc) (distance point (center arc)))))

(defun make-arc (center radius start-angle end-angle)
  (make-instance 'arc :center center :radius radius 
                 :start-angle start-angle :end-angle end-angle))

(defmethod intersect ((arc arc) (line line))
  (generic-intersect arc line))

(defmethod intersect ((line line) (arc arc))
  (generic-intersect arc line))

(defmethod intersect ((arc1 arc) (arc2 arc))
  (generic-intersect arc1 arc2))

(defclass convex ()
  ((sides :initarg :sides :accessor sides)
   (vertices :initarg :vertices :accessor vertices 
             :documentation "List of vertices that could be used
             by LOS routines.")))

(defmethod orient ((convex convex) point)
  (let ((sides (sides convex)))
    (flet ((op (side) (orient side point)))
      (if (op (first sides))
          (every #'op sides)
          (notany #'op sides)))))

(defun make-convex (sides vertices)
  (make-instance 'convex :sides sides :vertices vertices))

(defun make-cone (center radius start-angle end-angle)
  (let* ((arc (make-arc center radius start-angle end-angle))
         (start (start arc))
         (end (end arc)))
    (make-convex (list (make-line center start)
                       arc (make-line end center))
                 (list center start end))))

(defun make-convex-polygon (vertices)
  "Makes a convex polygon from list of vertices. Doesn't check if
the result is actually a convex polygon."
  (let ((sides (loop for ver on vertices
                     for v1 = (first ver)
                     for v2 = (second ver)
                     unless v2 do (setf v2 (first vertices))
                     collect (make-line v1 v2))))
    (make-convex sides vertices)))

(defmethod intersect-p (curve (convex convex))
  (loop for side in (sides convex)
       thereis (intersect-p curve side)))

(defmethod intersect (curve (convex convex))
  (loop for side in (sides convex)
        nconcing (intersect curve side)))

(defmethod intersect-p ((convex convex) curve)
  (loop for side in (sides convex)
       thereis (intersect-p curve side)))

(defmethod intersect ((convex convex) curve)
  (loop for side in (sides convex)
        nconcing (intersect curve side))) 

(defmethod intersect-p ((c1 convex) (c2 convex))
  (loop for side in (sides c1)
       thereis (intersect-p side c2)))

(defmethod intersect ((c1 convex) (c2 convex))
  (loop for side in (sides c1)
        nconcing (intersect side c2)))

(defgeneric subcurve (curve start end)
  (:documentation "Create a curve that is subcurve of curve with
  given start and end"))

(defmethod subcurve ((line line) start end)
  (make-line start end))

(defmethod subcurve ((arc arc) start end)
  (let ((center (center arc)))
    (make-arc center (radius arc) 
              (signum (- start center))
              (signum (- end center)))))

(defun cmppoints (pt1 pt2)
  "Returns true if points are closer than *epsilon* to each other"
  (< (abs (- pt1 pt2)) *epsilon*))

(defun revside (side)
  "Reverses the orientation of side"
  (subcurve side (end side) (start side)))

(defun sort-sides (sides &aux buffer)
  (unless sides (return-from sort-sides nil))
  (let ((curside (pop sides)))
    (push curside buffer)
    (dotimes (i (length sides) (reverse buffer))
      (let ((next (loop with mindist = 0.1 and mside
                        for side in sides
                        for dists = (abs (- (end curside) (start side)))
                        for diste = (abs (- (end curside) (end side)))
                        when (< dists mindist)
                        do (setf mindist dists) (setf mside side)
                        when (< diste mindist)
                        do (setf mindist diste) (setf mside side)
                        finally (return mside))))
        (unless next (error "Bad convex! Curside: ~a ~%~a" curside sides))
        (setf sides (remove next sides))
        (unless (cmppoints (end curside) (start next))
          (setf next (revside next)))
        (push next buffer) (setf curside next)))))

;;Possible bug?
;;(convex-intersection 
;; (make-convex-polygon (list #C(-1 -1) #C(-1 1) #C(1 1) #C(1 -1)))
;; (make-convex-polygon (list #C(-11.5 10) #C(10 11.5) 
;;                            #C(11.5 10) #C(10 -11.5))))     

(defun convex-intersection (c1 c2)
  (let (sides vertices)
    (flet ((sideloop (c1 c2)
             (loop for side in (sides c1)
                for ints = (intersect side c2)
                for len = (length ints)
                do (case len
                     (2 (destructuring-bind (a b) ints
                          (if (< (distance a (start side))
                                 (distance b (start side)))
                              (push (subcurve side a b) sides)
                              (push (subcurve side b a) sides))))
                     (1 (let ((a (car ints)) (s (start side)) (e (end side)))
                          (cond ((orient c2 s) 
                                 (push (subcurve side s a) sides))
                                ((orient c2 e)
                                 (push (subcurve side a e) sides)))))
                     (0 (when (and (orient c2 (start side)) 
                                   (orient c2 (end side)))
                          (push side sides)))))))
      (sideloop c1 c2)
      (sideloop c2 c1))
    ;;(setf sides (sort-sides sides))
     (handler-case (setf sides (sort-sides sides))
           ;(error () 
           ;  (setf *c1* c1 *c2* c2)
             ;(error "C1:~a~%C2:~a~%Sides:~a" (sides c1) (sides c2) sides))
             )
    (setf vertices (mapcar #'start sides))
    (when sides (make-convex sides vertices))))
             
(defclass base ()
  ((anchor :initarg :anchor :accessor anchor)
   (vector1 :initarg :vector1 :accessor vector1)
   (vector2 :initarg :vector2 :accessor vector2)))

(defun make-base (anchor vector1 vector2)
  (make-instance 'base :anchor anchor :vector1 vector1 :vector2 vector2))

(let (mem-base convex)
  (defun base-convex (base)
    (if (eql base mem-base)
        convex
        (let* ((p1 (anchor base))
               (p2 (+ p1 (vector1 base)))
               (s1 (make-line p1 p2))
               (p3 (+ p2 (vector2 base)))
               (s2 (make-line p2 p3))
               (p4 (+ p1 (vector2 base)))
               (s3 (make-line p3 p4))
               (s4 (make-line p4 p1))
               (sides (list s1 s2 s3 s4))
               (vertices (list p1 p2 p3 p4)))
          (setf mem-base base)
          (setf convex (make-instance 'convex :sides sides :vertices vertices))))))
        

(let (mem-base v11 v12 v22 det)       
  (defun base-coords (point base)
    (unless (eql base mem-base)
      (setf v11 (norm (vector1 base)))
      (setf v22 (norm (vector2 base)))
      (setf v12 (scalar-product (vector1 base) (vector2 base)))
      (setf det (- (* v11 v22) (* v12 v12)))
      (setf mem-base base))
    (let* ((z (- point (anchor base)))
           (x (/ (- (* (scalar-product z (vector1 base)) v22)
                    (* (scalar-product z (vector2 base)) v12))
                 det))
           (y (/ (- (* (scalar-product z (vector2 base)) v11)
                    (* (scalar-product z (vector1 base)) v12))
                 det)))
      (values x y))))

(defclass ray ()
  ((anchor :initarg :anchor :initarg :start 
           :accessor anchor :accessor start)
   (direction :initarg :direction :initarg :angle 
              :accessor direction :accessor angle)))

(defmethod print-object ((ray ray) s)
  (print-unreadable-object (ray s :type t :identity nil)
    (format s "~a ~a" (start ray) (direction ray))))

(defun make-ray (anchor direction)
  (make-instance 'ray :anchor anchor :direction direction))

(defun line-ray (line)
  (make-ray (start line) (- (end line) (start line))))
  
(defmethod end ((ray ray))
  (+ (start ray) (direction ray)))

(defmethod orient ((ray ray) point)
  (let ((v (- point (start ray))))
    (> (anti-product v (direction ray)) 0)))

(defmethod intersect-p ((ray ray) (line line))
  (let ((os (orient ray (start line))) (oe (orient ray (end line)))
        (oa (orient line (anchor ray))) 
        (od (orient line (+ (start line) (direction ray)))))
    (not (or (eql os oe) (eql oa od)))))

(defmethod intersect ((ray ray) (line line))
  (let ((s (start line)) (e (end line)) (a (anchor ray)) (d (direction ray)))
    (when (intersect-p ray line) 
      (let* ((v1 (- e s)) (v2 d) (z (- a s))
             (v11 (norm v1)) (v22 (norm v2))
             (v12 (scalar-product v1 v2))
             (det (- (* v11 v22) (* v12 v12))))
        (unless (< (abs det) *epsilon*)
          (list (+ s (* v1 (/ (- (* (scalar-product z v1) v22)
                                 (* (scalar-product z v2) v12))
                              det)))))))))

(defmethod intersect-p ((line line) (ray ray))
  (intersect-p ray line))

(defmethod intersect ((line line) (ray ray))
  (intersect ray line))

(defmethod intersect-p ((ray1 ray) (ray2 ray))
  (let ((oa1 (orient ray2 (anchor ray1))) 
        (od1 (orient ray2 (+ (anchor ray2) (direction ray1))))
        (oa2 (orient ray1 (anchor ray2))) 
        (od2 (orient ray1 (+ (anchor ray1) (direction ray2)))))
    (not (or (eql oa1 od1) (eql oa2 od2)))))

(defmethod intersect ((ray1 ray) (ray2 ray))
  (let ((a1 (anchor ray1)) (d1 (direction ray1)) 
        (a2 (anchor ray2)) (d2 (direction ray2)))
    (when (intersect-p ray1 ray2)
      (let* ((v1 d1) (v2 d2) (z (- a2 a1))
             (v11 (norm v1)) (v22 (norm v2))
             (v12 (scalar-product v1 v2))
             (det (- (* v11 v22) (* v12 v12))))
        (unless (< (abs det) *epsilon*)
          (list (+ a1 (* v1 (/ (- (* (scalar-product z v1) v22)
                                  (* (scalar-product z v2) v12))
                               det)))))))))

(defmethod distance ((point number) (ray ray))
  (let ((a (- point (start ray)))
        (b (direction ray)))
    (if (<= 0.0 (scalar-product a b))
        (abs (- a (project a b)))
        (distance point (start ray)))))

(defmethod distance ((ray ray) (point number))
  (distance point ray))

(defclass complex-figure ()
  ((figures :initarg :figures :accessor figures)))

(defmethod orient ((figure complex-figure) point)
  (loop for f in (figures figure)
       thereis (orient f point)))


(proclaim '(inline in-angle-p))
(defun in-angle-p (point anchor dir1 dir2)
  (and (> (anti-product (- point anchor) dir1) 0)
       (< (anti-product (- point anchor) dir2) 0)))

(defclass inf-cone (base) ())

(defmethod orient ((cone inf-cone) point)
  (in-angle-p point (anchor cone) (vector1 cone) (vector2 cone)))

(defun angular-in-p (side anchor dir1 dir2 bisector)
  (let ((p1 (start side))
        (p2 (end side)))
    (or (in-angle-p p1 anchor dir1 dir2)
        (in-angle-p p2 anchor dir1 dir2)
        (intersect-p bisector side))))
        
        
(defparameter *shadowcast-debug* nil)

(defun recursive-shadowcast (anchor dir1 dir2 walls)
  (when (< (abs (phase (/ dir1 dir2))) 0.01) 
    (return-from recursive-shadowcast nil))
  (when *shadowcast-debug* (format t "Anchor:~a dir1:~a dir2:~a~%Walls: ~a~%" anchor dir1 dir2 walls)) 
  (let* ((bisector (make-ray anchor (+ dir1 dir2)))
         (walls (remove-if-not (lambda (side)
                                 (angular-in-p side anchor dir1 dir2 bisector))
                              walls))
         (wall (let ((w (pop walls)))
                 (when *shadowcast-debug* (format t "Cur wall:~a Rem walls: ~a~%" w walls))
                 (unless w 
                   (return-from recursive-shadowcast
                     (list (make-instance 'inf-cone :anchor anchor
                                          :vector1 dir1 :vector2 dir2))))
                 w))
         (p1 (start wall)) (inp1 (in-angle-p p1 anchor dir1 dir2))
         (p2 (end wall)) (inp2 (in-angle-p p2 anchor dir1 dir2)))
    (flet ((bisection (p1 p2)
             (if (fast-orient anchor p1 p2)
               (let ((int (intersect (make-ray anchor dir2) wall)))
                 (when *shadowcast-debug* (format t "b1~%"))
                 (append (recursive-shadowcast anchor dir1 (- p1 anchor) walls)
                         (list (make-convex-polygon 
                                (list anchor p1 (if int (car int) p2))))))
               (let ((int (intersect (make-ray anchor dir1) wall)))
                 (when *shadowcast-debug* (format t "b2~%"))
                 (append (list (make-convex-polygon 
                                (list anchor (if int (car int) p2) p1)))
                         (recursive-shadowcast anchor (- p1 anchor) dir2 walls))))))
      (cond ((and inp1 inp2)
             (when *shadowcast-debug* (format t "inp1, inp2 ~%"))
             (if (in-angle-p p1 anchor dir1 (- p2 anchor))
                 (append (recursive-shadowcast anchor dir1 (- p1 anchor) walls)
                         (list (make-convex-polygon (list anchor p1 p2)))
                         (recursive-shadowcast anchor (- p2 anchor) dir2 walls))
                 (append (recursive-shadowcast anchor dir1 (- p2 anchor) walls)
                         (list (make-convex-polygon (list anchor p2 p1)))
                         (recursive-shadowcast anchor (- p1 anchor) dir2 walls))))
            (inp1 (when *shadowcast-debug* (format t "inp1 ~%")) (bisection p1 p2))
            (inp2 (when *shadowcast-debug* (format t "inp2 ~%")) (bisection p2 p1))
            (t (when *shadowcast-debug* (format t "none ~%"))
             (let ((int1 (intersect (make-ray anchor dir1) wall))
                     (int2 (intersect (make-ray anchor dir2) wall))
                     lp rp)
                 (if (fast-orient anchor p1 p2)
                     (setf lp p1 rp p2)
                     (setf lp p2 rp p1))
                 (list (make-convex-polygon (list anchor
                                                  (if int1 (car int1) lp)
                                                  (if int2 (car int2) rp))))))))))
                                     
(defun shadowcast (anchor dir1 dir2 blocks)
  (make-instance 'complex-figure :figures
                 (recursive-shadowcast anchor dir1 dir2
                                       (sort (loop for block in blocks
                                                   appending (sides block)) #'< 
                                                :key (lambda (side)
                                                       (distance anchor side))))))

(defun choose-target (targetlist &optional problist)
  "targetlist elements are ((x . y) . value)"
  (cond
    ((= (length targetlist) 0) nil)
    ;;    ((= (length targetlist) 1) (cdar targetlist))
    (t (attrset :cwhite)
       (mvprintw 10 62 "Choose target")
       (mvprintw 11 62 "p/-/left: previous")
       (mvprintw 12 62 "n/+/right: next")
       (mvprintw 13 62 "RET selects")
       (mvprintw 14 62 "ESC/SPC cancels")
     (unwind-protect
          (let ((idx 0) (len (length targetlist)) (cur (car targetlist)))
            (loop 
               (setf cur (nth idx targetlist))
               (put-char (caar cur) (cdar cur) (cons #\X :crose) 
                         :noscreen t :nomuffle t)
               (when problist
                 (clean-area 16 62 16 79)
                 (attrset :crose)
                 (mvprintw 16 62 (format nil "Chance to hit: ~2,'0d%"
                                         (floor (* (nth idx problist) 100)))))
               (case (getch)
                 ((261 54 454 110 226 465 43) (incf idx)
                  (when (>= idx len) (setf idx 0)))
                 ((260 52 452 112 167 464 45) (decf idx)
                  (when (< idx 0) (setf idx (1- len))))
                 ((13 459 114 170 98 168)
                  (out-char (caar cur) (cdar cur))
                  (return-from choose-target (cdr cur)))
                 ((27 32) (out-char (caar cur) (cdar cur)) 
                  (return-from choose-target nil)))
               (out-char (caar cur) (cdar cur))))
       (clean-area 10 62 16 79)))))
          
            

;;SECTION 3: Display

(defun print-char (y x char)
  (attrset (if *color-fix* *color-fix* (cdr char)))
  (mvaddch y x (car char)))

(defun refresh-screen ()
  (refresh))

(defun point-to-coords (point &key nonil)
  "Returns screen coordinates for the given point under screen base."
  (multiple-value-bind (x y) (base-coords point *screen-base*)
    (when (or nonil
              (and (< (- 0 *epsilon*) x (+ 1 *epsilon*)) 
                   (< (- 0 *epsilon*) y (+ 1 *epsilon*))))
      (let ((nx (if nonil (floor (* x *width*))
                    (max 0 (floor (* x *width*)))))
            (ny (if nonil (floor (* y *height*))
                    (max 0 (floor (* y *height*))))))
        (values (if (or nonil (< nx *width*)) nx (1- *width*))
                (if (or nonil (< ny *height*)) ny (1- *height*)))))))

(defun coords-to-point (x y)
  "Returns the point that has coordinates (x y) under screen base"
  (let ((basex (/ (+ x 0.5) *width*))
        (basey (/ (+ y 0.5) *height*)))
    (+ (anchor *screen-base*) 
       (* basex (vector1 *screen-base*))
       (* basey (vector2 *screen-base*)))))

(defun muffle-color (color)
  (case color
    (:clblue :cblue)
    (:cgreen :cdgreen)
    (:csky :ccyan)
    (:crose :cred)
    (:cpink :cpurple)
    (:cyellow :cbrown)
    ((:cdark :cwhite) :cgray)
    (t color)))

(defun muffle-char (char)
  (cons (car char) (muffle-color (cdr char))))
  
(defun put-char (x y char &key monster nomuffle noscreen)
  (when (and (<= 0 x (1- *width*)) (<= 0 y (1- *height*)))
    (if monster (setf (aref *monscreen* x y) monster)
        (unless noscreen (setf (aref *curscreen* x y) char)))
    (print-char (+ *dy* (- *height* y 1)) (+ *dx* x) 
                (if (and (not nomuffle) (zerop (bit *fovmap* x y)))
                    (muffle-char char) char))))

(defun move-cursor (x y)
  (when (and (<= 0 x (1- *width*)) (<= 0 y (1- *height*)))
    (move (+ *dy* (- *height* y 1)) (+ *dx* x))))


(defclass tchar ()
  ((object :initarg :object :initform nil :accessor object)
   (tag :initarg :tag :initarg :name :initform "" :accessor tag :accessor name)
   (flags :initarg :flags :initform nil :accessor flags)))

(defun make-tchar (object tag flags)
  (make-instance 'tchar :object object :tag tag :flags flags))

(defun put-tchar (x y tchar)
  (when (and (<= 0 x (1- *width*)) (<= 0 y (1- *height*)))
    (setf (aref *tagscreen* x y) tchar)))

(defparameter *default-char* (cons #\. :cgreen))
(defparameter *default-tchar* (make-instance 'tchar))

(defun clear-screen ()
  (attrset (muffle-color (cdr *default-char*)))
  (loop for i from 0 below *width* do
       (loop for j from 0 below *height* do
            (mvaddch (+ j *dy*) (+ i *dx*) (car *default-char*))
            (setf (aref *curscreen* i j) *default-char*)
            (setf (aref *monscreen* i j) nil)
            (setf (bit *fovmap* i j) 0)
            (put-tchar i j *default-tchar*))))

(defun clean-line (line)
  (loop for i from 0 to 79 do (mvaddch line i #\Space)))

(defun clean-area (y1 x1 y2 x2)
  (loop for y from y1 to y2
        do (loop for x from x1 to x2
                 do (mvaddch y x #\Space))))

(defun out-char (x y)
  "Update only the char on (x y)"
  (when (and (<= 0 x (1- *width*)) (<= 0 y (1- *height*)))
    (let ((mon (aref *monscreen* x y))
          (fov (bit *fovmap* x y)))
      (if (and mon (= fov 1)) (draw mon)
          (put-char x y (aref *curscreen* x y))))))
   
(defun update-monscreen (monster flag)
  (let ((x (x monster)) (y (y monster)))
    (when (and (<= 0 x (1- *width*)) (<= 0 y (1- *height*)))
      (setf (aref *monscreen* x y) (if flag monster nil))
      (out-char x y))))

(defun safe-print (y x str)
  "Workaround for a bug"
  (loop for c across str
        for xx = x then (1+ xx)
        do (mvaddch y xx c)))

(defun display-simple-message (str)
  (attrset :cdark)
  (clean-line 0)
  (safe-print 0 1 str)
  (push str *allmessages*)
  (refresh) "")

(defun yes-or-no (str)
  (display-simple-message (format nil "~a (y/n)" str))
  (case (getch)
    ((121 173 89 141) t)
    ((110 78 226 146) nil)))

(defun display-message (str)
  (if str 
      (progn
        (attrset :cdark)
        (clean-line 0)
        (safe-print 0 1 str)
        (when *messages*
          (attrset :cyellow)
          (mvprintw 0 70 "[MORE]"))
        (refresh))
      (display-simple-message "No more messages.")))

(defun message (str &rest args)
  (let ((s (apply #'format nil str args)))
    (push s *messages*) (push s *allmessages*)))

(defun merge-messages ()
  (loop with mstr = ""
        for m in *messages*
        if (<= (+ (length m) (length mstr)) 70)
        do (setf mstr (concatenate 'string mstr m " "))
        else collect mstr into mess
             and do (setf mstr (concatenate 'string m " "))
        finally (setf *messages* (nconc mess (list mstr)))))

(defun recap-messages ()
  (erase)
  (attrset :cgray)
  (loop for x = *allmessages* then (cdr x)
     for i from 1 to 20
     do (safe-print i 1 (if (car x) (car x) "-"))))

(defun visible-monsters ()
  (loop for x from 0 below *width* nconcing
       (loop for y from 0 below *height*
          for mon = (aref *monscreen* x y)
          for fov = (bit *fovmap* x y)
          when (and mon (not (eql mon *player*)) (= fov 1))
          collect (cons (cons x y) mon))))

(defun show-fov ()
  (loop for (_ . mon) in (visible-monsters)
       do (highlight-fov mon)))

;;   (loop for x from 0 below *width* do 
;;        (loop for y from 0 below *height*
;;           for mon = (aref *monscreen* x y)
;;           for fov = (bit *fovmap* x y)
;;           when (and mon (not (eql mon *player*)) (= fov 1))
;;           do (highlight-fov mon))))

(defun redraw-screen ()
  (erase)
  (draw-map-objects)
  (when *player* 
    (update-fov)
    (move-cursor (x *player*) (y *player*))
    (show-tag (x *player*) (y *player*))
    (show-status))
  (when *show-fov* (show-fov)))

(defun show-tag (x y)
  (attrset :cyellow)
  (clean-line 23)
  (let ((tag (tag (aref *tagscreen* x y))))
    (when tag
      (mvprintw 23 1 (format nil "~a" tag)))))

(defun change-base (anchor vector1 vector2)
  (setf *screen-base* (make-base anchor vector1 vector2))
  (draw-map-objects)
  ;;collect houses
  (setf *curwalls* nil)
  (loop for x from 0 below *width* do
       (loop for y from 0 below *height*
             for tag = (aref *tagscreen* x y)
             when (housep tag)
             do (pushnew tag *curwalls*)))
  (when *player* (update-fov))
  (when *show-fov* (show-fov))
  )

(defun update-fov ()
  (unless *player* (return-from update-fov))
  (let ((fov (shadowcast (position *player*)
                         (* (direction *player*) *fov-angle*)
                         (/ (direction *player*) *fov-angle*)
                         (mapcar #'area *curwalls*))))
    (loop for x from 0 below *width* do
         (loop for y from 0 below *height*
              for old = (bit *fovmap* x y)
              for pt = (coords-to-point x y)
              for new = (if (and (< (distance (position *player*) pt) 
                                    *sight-range*)
                                 (orient fov pt))
                                 1 0)
              unless (= old new)
              do (setf (bit *fovmap* x y) new)
                 (out-char x y)))
    (setf *curfov* fov)
    (setf (bit *fovmap* (x *player*) (y *player*)) 1)
    (out-char (x *player*) (y *player*))))
         
(defun highlight-fov (monster)
  (let ((fov (shadowcast (position monster)
                         (* (direction monster) *fov-angle*)
                         (/ (direction monster) *fov-angle*)
                         (mapcar #'area *curwalls*)))
        (*color-fix* :cyellow))
    (loop for x from 0 below *width* do
         (loop for y from 0 below *height*
               for pt = (coords-to-point x y) 
               for f = (and (< (distance (position monster) pt) *sight-range*)
                            (orient fov pt))
               when f do (out-char x y)))))
                                 
(defun show-status ()
  (clean-area 1 65 5 79)
  (attrset :crose)
  (mvprintw 1 65 (format nil "HP:~a/~a" (hp *player*) (max-hp *player*)))
  (attrset :cred)
  (mvprintw 2 65 (format nil "Kills:~d" *kill-count*))
  (attrset :cgray)
  (mvprintw 3 65 (format nil "Turns:~d" *turn-count*))
  (mvprintw 4 65 (format nil "Zoom:~@d" *zoom-level*))
  (attrset :cyellow)
  (if *show-fov* (mvprintw 5 65 "FOV On")))

;;SECTION 4: Mapobjects

;;todo: fill-convex, draw-line and similar routines

(defun fill-convex (convex char &optional tchar)
  (let (minx maxx miny maxy)
    (dolist (v (vertices convex))
      (multiple-value-bind (x y) (point-to-coords v :nonil t)
        (when (or (not minx) (< x minx)) (setf minx x))
        (when (or (not maxx) (> x maxx)) (setf maxx x))
        (when (or (not miny) (< y miny)) (setf miny y))
        (when (or (not maxy) (> y maxy)) (setf maxy y))))
    (when (and (>= maxx 0) (< minx *width*)
               (>= maxy 0) (< miny *height*))
      (loop for x from (max 0 minx) to (min (1- *width*) maxx)
         do (loop for y from (max 0 miny) to (min (1- *height*) maxy)
               when (orient convex (coords-to-point x y))
               do (put-char x y char)
                 (when tchar (put-tchar x y tchar)))))))

(defun tracer (y1 x1 y2 x2)
  "Adapted from The Sewer Massacre"
  (if (= x1 x2) (loop for y from (min y1 y2) to (max y1 y2)
                   collecting (cons y x1))
      (flet ((line (x)
               (/ (+ (* (- x x1 1/2) (+ y2 1/2))
                     (* (- x2 -1/2 x) (+ y1 1/2)))
                  (- x2 x1))))
        (let ((xmin (min (+ x1 1/2) (+ x2 1/2)))
              (xmax (max (+ x1 1/2) (+ x2 1/2))))
          (loop for x from xmin below xmax by 1/2
             for y* = (line x)
             for y** = (line (+ x 1/2))
             for ymin = (floor (min y* y**))
             for ymax = (if (= y* y**) (+ ymin 1) (max y* y**))
             nconcing (let ((l (loop for y from ymin below ymax
                                    collecting (cons y (floor x)))))
                        (if (< y** y*) (nreverse l) l))
                       into pts
             finally (return (remove-duplicates pts :test #'equal)))))))

(defun dir-keyword (dx dy)
  (ecase dx
    (1 (ecase dy (1 :ne) (0 :e) (-1 :se)))
    (0 (ecase dy (1 :n) (-1 :s)))
    (-1 (ecase dy (1 :nw) (0 :w) (-1 :sw)))))

(defun base-line-char (dir)
  (ecase dir
    (:none #\.)
    ((:ne :sw) #\/)
    ((:w :e) #\-)
    ((:nw :se) #\\)
    ((:n :s) #\|)))

(defun line-char (dir1 dir2)
  (cond ((eql dir1 :none) (base-line-char dir2))
        ((eql dir2 :none) (base-line-char dir1))
        ((eql dir1 dir2) (base-line-char dir1))
        ((eql dir1 :n)
         (ecase dir2 (:ne #\/) ((:sw :e) #\,) ((:se :s :w) #\.) (:nw #\\)))
        ((eql dir1 :ne)
         (ecase dir2 (:n #\;) (:e #\_) (:nw #\)) (:se #\_) ((:s :w :sw) #\,)))
        ((eql dir1 :e)
         (ecase dir2 ((:w :ne) #\-) (:se #\_) (:nw #\`) 
                (:n #\') (:sw #\,) (:s #\.)))
        ((eql dir1 :se)
         (ecase dir2 ((:w :nw :n :e) #\`) (:ne #\") (:s #\\) (:sw #\))))
        ((eql dir1 :s)
         (ecase dir2 (:nw #\`) ((:n :ne :w) #\') (:sw #\;)
                (:se #\:) (:e #\`)))
        ((eql dir1 :sw)
         (ecase dir2 ((:w :e) #\-) (:nw #\") ((:n :ne) #\')
                (:se #\() (:s #\:)))
        ((eql dir1 :w)
         (ecase dir2 (:nw #\`) (:n #\') ((:ne :w :se) #\-) ((:s :sw #\,))))
        ((eql dir1 :nw)
         (ecase dir2 ((:s :se) #\.) (:sw #\_) ((:w :e) #\-)
                (:n #\:) (:ne #\()))))
         
(defun draw-line (line color &optional tchar)
  (let* ((bc (base-convex *screen-base*))
         (ints (intersect line bc))
         (sline (case (length ints)
                  (2 (destructuring-bind (a b) ints
                       (subcurve line a b)))
                  (1 (let ((a (car ints)) (s (start line)) (e (end line)))
                       (cond ((orient bc s) (subcurve line s a))
                             ((orient bc e) (subcurve line s e)))))
                  (0 (when (and (orient bc (start line)) 
                                (orient bc (end line)))
                       line)))))
    (when sline
      (multiple-value-bind (sx sy) (point-to-coords (start sline) :nonil t)
        (multiple-value-bind (ex ey) (point-to-coords (end sline) :nonil t)
          ;;(format t "~a ~a (~a ~a ~a ~a) ~%" line sline sx sy ex ey)
          (let ((tr (tracer sx sy ex ey)))
            (loop with dir1 = :none and dir2
                  for ttr on tr
                  for (x . y) = (first ttr)
                  if (cdr ttr)
                  do (destructuring-bind (nx . ny) (second ttr)
                       (setf dir2 (dir-keyword (- nx x) (- ny y))))
                  else do (setf dir2 :none)
                  do (put-char x y (cons (line-char dir1 dir2) color))
                     (when tchar (put-tchar x y tchar))
                     (setf dir1 dir2))))))))

(defclass mapobject ()
  ((z-index :initarg :z-index :initform 0 
            :accessor z-index :type fixnum)))

(defgeneric draw (obj))

(defgeneric choose-point (obj)
  (:documentation "Choose a point in the object (for sector hashing)"))

(defclass street (mapobject edge)
  ((area :initarg :area :accessor area)
   (borders :initarg :borders :accessor borders)
   (z-index :initform 1)
   (dl :initarg :dl :accessor dl)
   (start :initarg :start :accessor start)
   (end :initarg :end :accessor end)))

(defmethod draw ((street street))
  (fill-convex (area street) (cons #\: :cdark) street)
  (draw-line (car (borders street)) :cgray street)
  (draw-line (cdr (borders street)) :cgray street)
  )

(defmethod choose-point ((street street))
  (/ (+ (start street) (end street)) 2))

(defclass street-intersection (mapobject)
  ((area :initarg :area :accessor area)
   (z-index :initform 0)
   (streets :initarg :streets :accessor streets)))

(defmethod draw ((int street-intersection))
  (fill-convex (area int) (cons #\: :cdark)))

(defmethod choose-point ((int street-intersection))
  (first (vertices (area int))))

(defmethod name ((int street-intersection))
  (format nil "~a" (remove-duplicates (mapcar #'name (streets int)) 
                                      :test #'equal)))

(defclass house (mapobject)
  ((area :initarg :area :accessor area)
   (name :initarg :name :accessor tag :accessor name)
   (color :initarg :color :initform :cdark :accessor color)
   (z-index :initform 3)
   ))

(defmethod draw ((house house))
  (fill-convex (area house) (cons #\# (color house)) house))

(defmethod choose-point ((house house))
  (first (vertices (area house))))

(defun occupied (x y &key excluding)
  (or (let ((mon (aref *monscreen* x y)))
        (if (eql mon excluding) nil
            mon))
      (housep (aref *tagscreen* x y))))

(defun housep (obj)
  (eql (class-of obj) (find-class 'house)))
  
(defclass corpse (mapobject)
  ((name :initarg :name :accessor name)
   (position :initarg :position :accessor position)
   (z-index :initform 2)))

(defmethod x ((corpse corpse))
  (floor (* (base-coords (position corpse) *screen-base*) *width*)))

(defmethod y ((corpse corpse))
  (floor (* (nth-value 1 (base-coords (position corpse) *screen-base*))
            *height*)))

(defmethod draw ((corpse corpse))
  (put-char (x corpse) (y corpse) (cons #\% :cred)))

(defmethod choose-point ((corpse corpse))
  (position corpse))

(defun create-corpse (monster)
  (let ((crp (make-instance 'corpse :name (name monster) 
                            :position (position monster))))
    (push crp *mapobjects*)
    (push crp (gethash (point-to-sector (position monster)) 
                          *map-sector-hash*))
    (push-nearby-objects crp)
    (draw crp) (refresh)))

(defun corpsep (obj)
  (eql (class-of obj) (find-class 'corpse)))

;;SECTION 5: City generator

(defvar *sprouts* nil "List of street sprouts. Each sprout is a ray object.")

(defvar *vertex-hash* (make-hash-table :test #'eql)
  "Maps the point to the list of edges that go to it")

(defvar *sector-hash* (make-hash-table :test #'equal)
  "Lists all vertices in the sector and the neighboring sectors")

(defvar *chains* nil)

(defparameter *sector-size* 0.8)

(defparameter *lock-threshold* 0.1)

(defparameter *withering-chance* 0.1)
(defparameter *withering-threshold* 10)

(defparameter *standard-length* 0.3)
(defparameter *standard-length-chance* 0.75)
(defparameter *min-length* 0.2)
(defparameter *max-length* 0.5)
(defparameter *min-distance* 0.1)

(defparameter *standard-cross-chance* 0.75)
(defparameter *half-cross-chance* 0.75)

(defparameter *min-angle* 1.0 "in radians")
(defparameter *minmin-angle* 0.5 "for locked connections")

(defparameter *street-width* 0.05)

(defparameter *house-tries* 4)

(defparameter *min-house-width* 0.1)
(defparameter *max-house-width* 0.2)

(defparameter *min-house-depth* 0.05)
(defparameter *max-house-depth* 0.15)

(defparameter *min-enemy-distance* 1.0)

(defclass edge (line)
  ((tag :initarg :tag :initarg :name 
        :accessor name :accessor tag)
   (leftp :initform nil :accessor leftp)
   (rightp :initform nil :accessor rightp)))

(defun make-edge (start end tag)
  (make-instance 'edge :start start :end end :tag tag))

(defclass sprout (ray)
  ((tag :initarg :tag :initarg :name 
        :accessor name :accessor tag)))

(defun make-sprout (start direction tag)
  (make-instance 'sprout :start start :direction direction :tag tag))

(defun point-to-sector (pt)
  (cons (floor (/ (x pt) *sector-size*)) (floor (/ (y pt) *sector-size*))))

(defun neighbors (sector)
  (let (res) 
    (loop for i from -1 to 1 do 
         (loop for j from -1 to 1 do
              (push (cons (+ i (car sector)) (+ j (cdr sector))) res)))
    res))

(defun line-end-p (point line)
  (or (= point (start line)) (= point (end line))))

(defun validate-line (line vertices)
  (and
   (loop for edge in (gethash (end line) *vertex-hash*)
         for other-end = (if (eql (start edge) (end line))
                             (end edge) (start edge))
        always (> (abs (phase (/ (- other-end (end line))
                                 (- (start line) (end line)))))
                  *minmin-angle*))
   (loop for v in vertices
      always (or (line-end-p v line) 
                 (loop for edge in (gethash v *vertex-hash*)
                    always (or (line-end-p (start line) edge)
                               (line-end-p (end line) edge)
                               (and (not (intersect-p line edge))
                                    (> (distance (end line) edge)
                                       *min-distance*))))))))

(defun same-line-p (line1 line2)
  (or (and (= (start line1) (start line2))
           (= (end line1) (end line2)))
      (and (= (start line1) (end line2))
           (= (end line1) (start line2)))))


(defun register-edge (line)
  (pushnew line (gethash (start line) *vertex-hash*) :test #'same-line-p)
  (pushnew line (gethash (end line) *vertex-hash*) :test #'same-line-p))

(defun register-vertex (point)
  (loop for sector in (neighbors (point-to-sector point))
     do (push point (gethash sector *sector-hash*))))

(defun remove-sprout (sprout)
  (setf *sprouts* (delete sprout *sprouts*)))

(defun random-length ()
  (if (< (random 1.0) *standard-length-chance*)
      *standard-length*
      (+ *min-length* (random (- *max-length* *min-length*)))))

(defun generate-tag ()
  ;;Placeholder...
  (format nil "~a street" (random 1000)))

(defun random-dir ()
  (loop for c = (complex (1- (random 2.0)) (1- (random 2.0)))
        while (< (abs c) *epsilon*)
        finally (return (signum c))))


(defun validate-dirs (dirlist)
  (loop for dir1 in dirlist
     always (loop for dir2 in dirlist
                 always (or (eql dir2 dir1)
                            (> (abs (phase (/ dir1 dir2))) *min-angle*)))))

(defun generate-intersection (d)
  (loop for dirs = (loop for i from 1 to (+ 2 (random 3))
                        collect (random-dir))
       until (validate-dirs (cons (- d) dirs))
       finally (return dirs)))

(defun validate-position (pt blocks)
  ;;(notany (lambda (b) (orient b pt)) blocks)
  (loop for b in blocks
       when (orient b pt) do (return nil)
       minimizing (distance pt (car (vertices b))) into mind
       finally (return (< mind *min-enemy-distance*))))
         
(defun run-sprout (sprout)
  ;;Random withering chance
  (when (and (> (length *sprouts*) *withering-threshold*)
             (< (random 1.0) *withering-chance*))
    (remove-sprout sprout)
    (return-from run-sprout nil))
  (let* ((s (start sprout)) (d (direction sprout))
         (nearby-vertices (gethash (point-to-sector s) *sector-hash*))
         lockv)
    
    ;;Crowdy vertex check
    (unless (loop for edge in (gethash s *vertex-hash*)
                  for other-end = (if (eql (start edge) s) (end edge) (start edge))
                  always (> (abs (phase (/ (- other-end s) d)))
                            *minmin-angle*))
      (remove-sprout sprout)
      (return-from run-sprout nil))

    (when nearby-vertices
      (loop with mindist = (* 3 *sector-size*)
         for v in nearby-vertices
         for dist1 = (distance v sprout)
         for dist2 = (distance v s)
         when (and (< dist1 *lock-threshold*)
                   (< dist2 mindist)
                   (/= v s))
         do (setf mindist dist2) (setf lockv v)))
    (when lockv
      (let ((lock-line (make-edge s lockv (tag sprout))))
        (when (validate-line lock-line nearby-vertices)
          (register-edge lock-line)
          (setf (start sprout) lockv) ;;Chances are the street continues
          (return-from run-sprout lock-line))))
    (let* ((length (random-length))
           (e (+ s (* length (signum d))))
           (edge (make-edge s e (tag sprout))))
      (when (validate-line edge nearby-vertices)
        (register-vertex e)
        (register-edge edge)
        (if (< (random 1.0) *standard-cross-chance*)
            (let ((tag (generate-tag)))
              (setf (start sprout) e)
              (push (make-sprout e (* d #C(0 1)) tag) *sprouts*)
              (push (make-sprout e (* d #C(0 -1)) tag) *sprouts*))
            (if (< (random 1.0) *half-cross-chance*)
                (case (random 3)
                  (0 (setf (start sprout) e)
                     (push (make-sprout e (* d #C(0 1)) (generate-tag)) *sprouts*))
                  (1 (setf (start sprout) e)
                     (push (make-sprout e (* d #C(0 -1)) (generate-tag)) *sprouts*))
                  (2 (remove-sprout sprout)
                     (let ((tag (generate-tag)))
                       (push (make-sprout e (* d #C(0 1)) tag) *sprouts*)
                       (push (make-sprout e (* d #C(0 -1)) tag) *sprouts*))))
                (progn (remove-sprout sprout)
                       (loop for dir in (generate-intersection d)
                          do (push (make-sprout e dir (generate-tag)) *sprouts*))))) 
        (return-from run-sprout edge)))
    ;;If everything failed...
    (remove-sprout sprout)))

(defun random-list (list)
  (nth (random (length list)) list))
    
(defun generate-street-map (n)
  (setf *sprouts* nil)
  (setf *vertex-hash* (make-hash-table :test #'eql))
  (setf *sector-hash* (make-hash-table :test #'equal))
  (let ((start #C(0.0 0.0))
        (dir (random-dir)))
    (register-vertex start)
    (push (make-sprout start dir (generate-tag)) *sprouts*)
    (clean-line 23) 
    (display-simple-message "Please wait while the map is generated...")
    (refresh) (attrset :cyellow)
    (loop for i from 1 to n
       when (zerop (mod i 50)) do (mvaddch 23 (/ i 50) #\+) (refresh) 
       do (run-sprout (random-list *sprouts*)))))
    
(defun draw-street-map ()
  "Debug only"
  (clear-screen)
  (loop for lines being each hash-value in *vertex-hash*
         do (loop for line in lines
               do (draw-line line :cred)))
  (refresh-screen))

(defun interactive-street-map ()
  (let ((step (/ (vector1 *screen-base*) 5)))
    (loop (let ((a (anchor *screen-base*))
                (v1 (vector1 *screen-base*))
                (v2 (vector2 *screen-base*)))
            (case (getch)
              ((261 54 454) (setf *screen-base* 
                                  (make-base (+ a step) v1 v2)))
              ((260 52 452) (setf *screen-base* 
                                  (make-base (+ a (* step -1)) v1 v2)))
              ((259 56 450) (setf *screen-base* 
                                  (make-base (+ a (* step #C(0 1))) v1 v2)))
              ((258 50 456) (setf *screen-base* 
                                  (make-base (+ a (* step #C(0 -1))) v1 v2)))
              ((51 457) (setf *screen-base* 
                              (make-base (+ a (* step #C(1 -1))) v1 v2)))
              ((49 455) (setf *screen-base* 
                              (make-base (+ a (* step #C(-1 -1))) v1 v2)))
              ((57 451) (setf *screen-base* 
                              (make-base (+ a (* step #C(1 1))) v1 v2)))
              ((55 449) (setf *screen-base* 
                              (make-base (+ a (* step #C(-1 1))) v1 v2)))
              ((464) (setf *screen-base*
                           (make-base (- a (/ v1 2) (/ v2 2))
                                      (* v1 2) (* v2 2))
                           step (* step 2)))
              ((465) (setf *screen-base*
                           (make-base (+ a (/ v1 4) (/ v2 4))
                                      (/ v1 2) (/ v2 2))
                           step (/ step 2)))
              ((27 113 169 108 164 13 32) (return))))
            (draw-street-map))))

(defmethod update-instance-for-different-class ((edge edge) (new street) &key)
  (let* ((s (start edge)) (e (end edge))
         (dl (* (signum (- e s)) (complex 0 (/ *street-width* 2))))
         (s+ (+ s dl)) (s- (- s dl)) (e+ (+ e dl)) (e- (- e dl)))
    (setf (area new) (make-convex-polygon (list s+ e+ e- s-))
          (dl new) dl
          (z-index new) 1
          (name new) (tag edge)
          (borders new) (cons (make-line s+ e+) (make-line s- e-)))))

(defun validate-house (house chain other-houses)
  ;;(format t "~a: ~a ~%" house (mapcar #'car other-houses)) 
  (and (loop for (side . orient) in chain
          for brd = (if (eql orient :right) 
                        (car (borders side))
                        (cdr (borders side)))
             for md = (loop for v in (vertices house)
                           minimizing (distance v brd))
          never (or (< (* 2 md) *street-width*) 
                    (intersect-p house brd)))
       (loop for (h . nil) in other-houses
          never (intersect-p house h))))
  

(defun edge-dir (point edge)
  (if (= (start edge) point) 
      (- (end edge) point)
      (- (start edge) point)))

(defun nextedge (point edge)
  (let* ((other-edges (remove edge (gethash point *vertex-hash*)))
         (dir (- (edge-dir point edge)))
         (dirs (mapcar (lambda (e) (cons (/ (edge-dir point e) dir) e))
                       other-edges)))
    (if other-edges
        (cdr (first (sort dirs #'< :key (lambda (d) (phase (car d))))))
        edge)))

(defun walk-edge (start edge)
  (let ((next (if (= start (start edge)) (end edge) (start edge))))
    (list next (nextedge next edge))))

(defun traverse-chain (start edge)
  (loop for (nextstart nextedge) = (list start edge)
        then (walk-edge nextstart nextedge)
        for orient = (if (= nextstart (start nextedge)) :left :right)
        for i from 1
        until (and (> i 1) (eql nextstart start) (eql nextedge edge))
        when (eql orient :left) do (setf (leftp nextedge) t)
        when (eql orient :right) do (setf (rightp nextedge) t)
        collecting (cons nextedge orient)
       ;;finally (print or)
       ))
          
  

(defun process-streets ()
  (setf *mapobjects* nil)
  (setf *map-sector-hash* (make-hash-table :test #'equal))
  (let (all-edges all-vertices all-blocks)
    (loop for edges being each hash-value in *vertex-hash*
          using (hash-key v)
          do (push v all-vertices)
         (dolist (edge edges)
           (pushnew edge all-edges :test #'same-line-p)))
    (dolist (edge all-edges) 
      (change-class edge 'street)
      (push edge *mapobjects*))
    (loop for v in all-vertices
          for directions = (loop for edge in (gethash v *vertex-hash*)
                                 when (= (start edge) v)
                                 collect (list (end edge)
                                               (- (start edge) (end edge))
                                               edge)
                                 else collect (list (start edge)
                                                    (- (end edge) (start edge))
                                                    edge))
          for dirsorted = (sort directions #'< :key (lambda (k) 
                                                      (phase (second k))))
       when (second dirsorted)
       do (loop for dd on dirsorted
                for dir1 = (first dd)
                for dir2 = (if (second dd) (second dd) (first dirsorted))
                for dl1 = (* (signum (second dir1))
                             (complex 0 (/ *street-width* -2)))
                for dl2 = (* (signum (second dir2))
                             (complex 0 (/ *street-width* 2)))
                for c = (car (intersect (make-ray (+ (first dir1) dl1)
                                                  (second dir1))
                                        (make-ray (+ (first dir2) dl2)
                                                  (second dir2))))
                for edge1 = (third dir1) for edge2 = (third dir2)
                when c collect c into corners
                and do 
                  (cond ((= (end edge1) v)
                         (setf (end (cdr (borders edge1))) c))
                        ((= (start edge1) v)
                         (setf (start (car (borders edge1))) c)))
                  (cond ((= (end edge2) v)
                         (setf (end (car (borders edge2))) c))
                        ((= (start edge2) v)
                         (setf (start (cdr (borders edge2))) c)))
                else collect (+ (first dir1) (second dir1) dl1) into corners
                finally (when (>= (length corners) 3)
                          (push (make-instance 'street-intersection
                                               :area (make-convex-polygon corners)
                                               :streets (gethash v *vertex-hash*))
                                *mapobjects*))))
    (loop for m in *mapobjects*
       do (push m (gethash (point-to-sector (choose-point m))
                           *map-sector-hash*))
       when (typep m 'street)
       do (setf (area m) (make-convex-polygon (list (start (car (borders m)))
                                                    (end (car (borders m)))
                                                    (end (cdr (borders m)))
                                                    (start (cdr (borders m)))))))
    ;;Chain detection
    (setf *chains* nil)
    (loop for v in all-vertices
          do (loop for edge in (gethash v *vertex-hash*)
                  when (and (= v (start edge)) (not (leftp edge)))
                  do (push (traverse-chain v edge) *chains*)
                  when (and (= v (end edge)) (not (rightp edge)))
                  do (push (traverse-chain v edge) *chains*)))
  
    ;;Housing
    (attrset :cpink)
    (loop for chain in *chains*
         for i from 1
         for col = (random-list *colors*)
         when (zerop (mod i 20)) do (addch #\+) (refresh)
         do (loop with chain-houses 
                 for (side . orient) in chain
                 do (loop for i from 1 to *house-tries*
                          for width = (+ *min-house-width* 
                                         (random (- *max-house-width* 
                                                    *min-house-width*)))
                          for depth = (+ *min-house-depth*
                                         (random (- *max-house-depth*
                                                    *min-house-depth*)))
                          for brd = (if (eql orient :right) 
                                        (car (borders side))
                                        (cdr (borders side)))
                          for dl = (if (eql orient :right) 
                                        (dl side) (- (dl side)))
                          for dir = (- (end brd) (start brd))
                          for wn = (+ width (* 2 *street-width*))
                          when (> (abs dir) wn)
                          do (let* ((k (random (- 1.0 (/ wn (abs dir)))))
                                    (s (+ (start brd) dl (* k dir)
                                          (* *street-width* (signum dir))))
                                    (w (* (signum dir) width))
                                    (d (* (signum dl) depth))
                                    (house (make-convex-polygon (list s (+ s d) 
                                                                      (+ s d w)
                                                                      (+ s w)))))
                               (when (validate-house house chain chain-houses)
                                 (push (cons house side) chain-houses))))
                 finally (loop for (h . side) in chain-houses
                               for house = (make-instance 'house 
                                                          :area h
                                                          ;;:color col
                                                          :name (tag side))
                               ;;do (format t "~a ~%" (sides h))
                               do (push house *mapobjects*)
                                  (push house (gethash 
                                               (point-to-sector 
                                                (choose-point house))
                                               *map-sector-hash*))
                                  (push h all-blocks))))
    ;;Boundaries
    (loop for v in all-vertices
         minimizing (x v) into minx
         maximizing (x v) into maxx
         minimizing (y v) into miny
         maximizing (y v) into maxy
         finally (setf *map-minx* minx *map-miny* miny
                       *map-maxx* maxx *map-maxy* maxy))
    ;;Enemies
    (loop for i from 1 to *enemy-count*
          do (loop for rp = (complex (+ *map-minx* 
                                        (random (- *map-maxx* *map-minx*)))
                                     (+ *map-miny* 
                                        (random (- *map-maxy* *map-miny*))))
                  until (validate-position rp all-blocks)
                  finally (push (make-instance 'enemy :position rp
                                               :direction (random-dir))
                                *monsters*)))
    )
  (attrset :cgreen)
  (addstr "Ready")
  )


(let (membase result)
  (defun nearby-objects ()
    (if (eql *screen-base* membase) result
        (let* ((a (anchor *screen-base*)) 
               (v1 (vector1 *screen-base*)) (sv1 (* (signum v1) *sector-size* 1/2))
               (v2 (vector2 *screen-base*)) (sv2 (* (signum v2) *sector-size* 1/2))
               (p1 (- a sv1 sv2)) (p2 (+ a v1 sv1 (- sv2)))
               (p3 (+ a v1 v2 sv1 sv2)) (p4 (+ a v2 sv2 (- sv1)))
               (secs (mapcar #'point-to-sector (list p1 p2 p3 p4)))
               (xs (mapcar #'car secs)) (ys (mapcar #'cdr secs))
               (minx (apply #'min xs)) (maxx (apply #'max xs))
               (miny (apply #'min ys)) (maxy (apply #'max ys)))
          (setf membase *screen-base*)
          (setf result
                (loop for x from minx to maxx nconcing
                     (loop for y from miny to maxy
                        appending (gethash (cons x y) *map-sector-hash*)))))))
  (defun push-nearby-objects (object)
    (pushnew object result)))
    
                                 
(defun draw-map-objects (&key always-draw-monsters
                         &aux (large-scale (> (abs (vector1 *screen-base*)) 1.0)))
  (clear-screen)
  (loop for m in (sort (nearby-objects) #'< :key #'z-index) 
     unless (and large-scale (corpsep m))
     do (draw m))
  (loop for m in *monsters*
        for (x y) = (multiple-value-list (point-to-coords (position m))) 
     when always-draw-monsters
       do (when x (draw m))
     else
     unless (or (not x)
                (zerop (bit *fovmap* x y)))
     do (draw m)
     else do (when x (setf (aref *monscreen* x y) m)))
  (if *player* (draw *player*))
  (refresh-screen))
          
(defun interactive-map ()
  (let ((step (/ (vector1 *screen-base*) 5)))
    (loop (let ((a (anchor *screen-base*))
                (v1 (vector1 *screen-base*))
                (v2 (vector2 *screen-base*)))
            (case (getch)
              ((261 54 454) (setf *screen-base* 
                                  (make-base (+ a step) v1 v2)))
              ((260 52 452) (setf *screen-base* 
                                  (make-base (+ a (* step -1)) v1 v2)))
              ((259 56 450) (setf *screen-base* 
                                  (make-base (+ a (* step #C(0 1))) v1 v2)))
              ((258 50 456) (setf *screen-base* 
                                  (make-base (+ a (* step #C(0 -1))) v1 v2)))
              ((51 457) (setf *screen-base* 
                              (make-base (+ a (* step #C(1 -1))) v1 v2)))
              ((49 455) (setf *screen-base* 
                              (make-base (+ a (* step #C(-1 -1))) v1 v2)))
              ((57 451) (setf *screen-base* 
                              (make-base (+ a (* step #C(1 1))) v1 v2)))
              ((55 449) (setf *screen-base* 
                              (make-base (+ a (* step #C(-1 1))) v1 v2)))
              ((464) (setf *screen-base*
                           (make-base (- a (/ v1 2) (/ v2 2))
                                      (* v1 2) (* v2 2))
                           step (* step 2)))
              ((465) (setf *screen-base*
                           (make-base (+ a (/ v1 4) (/ v2 4))
                                      (/ v1 2) (/ v2 2))
                           step (/ step 2)))
              ((27 113 169 108 164 13 32) (return))))
            (draw-map-objects :always-draw-monsters t))))

(defun test-run ()
  (connect-console)
  (setf *screen-base* (make-base #C(0.0 0.0) #C(0.6 0.0) #C(0.0 0.3)))
  (generate-street-map 500)
  (process-streets)
  (clear-screen) (draw-map-objects)
  (refresh-screen)
  (interactive-map))

(defun left-side (street)
  (let ((brd (cdr (borders street))))
    (make-convex-polygon (list (start brd) (end brd) 
                               (end street) (start street)))))

(defun right-side (street)
  (let ((brd (car (borders street))))
    (make-convex-polygon (list (start brd) (end brd) 
                               (end street) (start street)))))

(defun visualise-chain (chain color)
  (loop for (street . orient) in chain
       do (fill-convex (if (eql orient :left) 
                           (left-side street)
                           (right-side street))
                       (cons #\* color)))
  (refresh-screen))

;;SECTION 6: Monsters

(defgeneric ai (monster)
  (:documentation "When called, returns an action that monster
  attempts to do."))

(defgeneric do-action (monster action)
  (:documentation "Execute an action"))

(defgeneric kill (monster &optional quitflag)
  (:documentation "Kill teh monster!"))

(defgeneric react (monster event)
  (:documentation "How a monster would react to an event.")
  (:method (m e) (declare (ignore m e)) nil))

(defclass monster ()
  ((name :initform "monster" :initarg :name :accessor name)
;;   (x :initform 0 :initarg :x :accessor x) 
;;   (y :initform 0 :initarg :y :accessor y)
   (position :initarg :position :accessor position)
   (direction :initarg :direction :accessor direction)
   (hp :initform 0 :initarg :hp :accessor hp) 
   (max-hp :initarg :max-hp :accessor max-hp)
   (monster-char :initarg :char :accessor monster-char))) 

(defmethod initialize-instance :after ((monster monster) &key)
  (setf (hp monster) (max-hp monster)))

(defmethod x ((monster monster))
  (floor (* (base-coords (position monster) *screen-base*) *width*)))

(defmethod y ((monster monster))
  (floor (* (nth-value 1 (base-coords (position monster) *screen-base*))
            *height*)))

(defmethod draw ((monster monster))
  (put-char (x monster) (y monster) (monster-char monster) :monster monster))

(defmethod kill ((monster monster) &optional quitflag)
  (declare (ignore quitflag))
  (cast (make-instance 'e-dead :monster monster))
  (setf *monsters* (delete monster *monsters*))
  ;;(setf (gethash (point-to-sector (position monster)) *monster-sector-hash*)
  ;;      (delete monster (gethash (point-to-sector (position monster)) 
  ;;                               *monster-sector-hash*)))
  (update-monscreen monster nil)
  (create-corpse monster)
  (redraw-screen)
  (when (and (eql *curmonster* *player*)
             (not (eql monster *player*)))
    (incf *kill-count*)))

;;  (cast (make-instance 'e-dead :monster monster))
;;  (setf *monsterstack* (delete monster *monsterstack* :key #'car))
;;  (setf (monster (aref *monscreen* (y monster) (x monster))) nil)
;;  (maybe-out-char (aref *map* (y monster) (x monster))))



(defclass player (monster) 
  ((name :initform "you")
   (max-hp :initform 10)
   (monster-char :initform (cons #\@ :cwhite))))

(defmethod ai :before (monster)
  (declare (ignore monster))
  (setf *cancelled* nil))

(defmethod ai :before ((monster player))
  (show-tag (x *player*) (y *player*))
  (show-status)
  (when (and *show-fov* (visible-monsters)) (redraw-screen))
  (if *zero-time-action* (setf *zero-time-action* nil))
  (progn (setf *messages* (nreverse *messages*))
         (merge-messages)
         (if *messages* 
             (display-message (pop *messages*))
             (unless *zero-time-action* (clean-line 0)))))

(defmethod ai ((monster player))
  (move-cursor (x monster) (y monster)) (refresh-screen)
  (let ((key (ubergetch)))
    (multiple-value-bind (action present-p) (gethash key *controls*)
      (if present-p 
          (progn (if (zerop (time action)) (setf *zero-time-action* t))
                 action) 
          (make-instance 'no-action :key key)))))

(defmethod ai :after ((monster player))
  (unless *zero-time-action* (setf *messages* nil) (incf *turn-count*))
  (when (or (> (x (position monster)) *map-maxx*)
            (< (x (position monster)) *map-minx*)
            (> (y (position monster)) *map-maxy*)
            (< (y (position monster)) *map-miny*))
    (display-simple-message "You have escaped The City!")
    (getch)
    (setf *quitflag* :escape)))

(defmethod kill ((monster player) &optional (quitflag :kill))
  (setf *quitflag* quitflag)
  (getch)
  (call-next-method))

(defun collect-monsters ()
  (loop with monsters 
     for x from 0 below *width*
     do (loop for y from 0 below *height*
              for m = (aref *monscreen* x y)
             when m do (pushnew m monsters))
     finally (return monsters)))

(defun not-seen (x y)
  (loop for m in (collect-monsters)
       never (aware-p m (coords-to-point x y))))
  

(defun generate-player ()
  ;;(repl-console 15)
  ;;(format t "~a ~%" *tagscreen*)
  (loop with x = (floor (/ *width* 2))
        and y = (floor (/ *height* 2))
        and phase = 2 and cnt = 1 and len = 1
        and dx = 0 and dy = 1
       while (or (occupied x y)
                 (not (not-seen x y)))
       do (incf x dx)
          (incf y dy)
          (decf cnt)
          (when (zerop cnt) (decf phase) (setf cnt len)
                (psetf dx (- dy) dy dx))
          (when (zerop phase) (incf len) (setf phase 2)
                (setf cnt len) (psetf dx (- dy) dy dx))
     finally (setf *player* 
                   (make-instance 'player :position (coords-to-point x y)
                                  :direction (random-dir)))
       (push *player* *monsters*)
       ;;(push *player* (gethash (point-to-sector (position *player*))
       ;;                        *monster-sector-hash*))
       ))


(defmethod react ((monster player) event)
  (when (interesting event) 
    (message (description event))))

(defmethod react ((monster player) (event e-step))
  (when (and (typep (monster event) 'enemy) 
             (< (random 1.0) *hearing-chance*))
    (multiple-value-bind (x y) (point-to-coords (position (monster event)))
      (when x
        (setf (bit *fovmap* x y) 1)
        (draw (monster event))))))

(defclass enemy (monster)
  ((name :initform "enemy")
   (max-hp :initform 5)
   (monster-char :initform (cons #\@ :cred))
   (state :initform :neutral :accessor state)
   (target :initform nil :accessor target)
   (last-seen :initform nil :accessor last-seen)
   (last-seen-from :initform nil :accessor last-seen-from)))

(defmethod ai ((monster enemy))
  (case (state monster)
    (:neutral
     (if (aware-p monster (position *player*))
         (progn (anger monster *player*)
                (make-instance 'turn :angle (/ (- (last-seen monster)
                                                  (position monster))
                                               (direction monster))))
         (make-instance 'wait)))
    (:hunt
     (if (aware-p monster (position *player*))
         (setf (state monster) :hostile
               (target monster) *player*
               (last-seen monster) (position *player*))
         (setf (state monster) :hunt2))
     (make-instance 'turn :angle (/ (- (last-seen monster)
                                       (position monster))
                                    (direction monster))))
    (:hunt2
     (if (aware-p monster (position *player*))
         (progn (setf (state monster) :hostile
                      (target monster) *player*
                      (last-seen monster) (position *player*))
                (make-instance 'turn :angle (/ (- (last-seen monster)
                                                  (position monster))
                                               (direction monster))))
         (if (< (distance (position monster) (last-seen-from monster))
                *moving-distance*)
             (make-instance 'turn :angle (/ (- (last-seen monster)
                                               (position monster))
                                            (direction monster)))
             (progn
               (do-action monster 
                 (make-instance 'turn 
                                :angle (/ (- (last-seen-from monster)
                                             (position monster))
                                          (direction monster))))
               (make-instance 'move-forward)))))
    (:hostile
     (let* ((pos (position (target monster))))
       (if (aware-p monster pos)
           (progn 
             (setf (last-seen monster) pos)
             (make-instance 'attack :monster (target monster) :weapon :rifle))
           (if (< (distance (position monster) (last-seen monster))
                  *moving-distance*)
               (progn (setf (state monster) :neutral)
                      (make-instance 'wait))
               (progn
                 (do-action monster 
                   (make-instance 'turn 
                                  :angle (/ (- (last-seen monster)
                                               (position monster))
                                            (direction monster))))
             (make-instance 'move-forward))))))))

(defun anger (monster target)
  (unless (eql (state monster) :hostile)
    (setf (state monster) :hostile 
          (target monster) target
          (last-seen monster) (position target))
    (when (eql (state monster) :neutral)
      (cast (make-instance 'e-m-angry :monster monster)))
    (cast (make-instance 'e-shout :monster monster :pt (position target)))))

;;These should be tweaked to allow sneaking

(defmethod react ((monster enemy) (event e-step))
  (when (and (typep (monster event) 'player) 
             (< (random 1.0) *hearing-chance*))
    (unless (and (eql (state monster) :hostile)
                 (aware-p monster (position *player*)))
      (do-action monster 
        (make-instance 'turn 
                     :angle (/ (- (pt event)
                                  (position monster))
                               (direction monster))))
      (anger monster (monster event)))))

(defmethod react ((monster enemy) (event e-combat))
  (when (eql (state monster) :neutral)
    (setf (state monster) :hunt))
  (let ((other-mon (if (eql (attacker event) *player*)
                       (defender event)
                       (attacker event))))
    (setf (last-seen monster) (position *player*)
          (last-seen-from monster) (position other-mon))))

(defmethod react ((monster enemy) (event e-shout))
  (when (eql (state monster) :neutral)
    (setf (state monster) :hunt))
  (setf (last-seen monster) (pt event)
        (last-seen-from monster) (position (monster event))))

;;(defmethod react ((monster enemy) (event e-wait))
;;  (when (typep (monster event) 'player)
;;    (anger monster (monster event))))


;;SECTION 7: Actions

(defclass action ()
  ((time :initform 5 :initarg :time :accessor time)))

(defun instead-do-action (act1 monster act2)
  (do-action monster act2)
  (setf (time act1) (time act2)))

(defun cancel-action (action &optional message)
  (when (and message (eql *curmonster* *player*))
    (message message))
  (setf (time action) 0 *cancelled* t *zero-time-action* t))

(defclass move-to (action)
  ((dp :initarg :dp :accessor dp)))

(defparameter *move-time-scale* 100.0)

(defmethod initialize-instance :after ((self move-to) &key)
  (setf (time self) (ceiling (abs (* (dp self) *move-time-scale*)))))

(defmethod do-action ((monster monster) (move-to move-to))
  (let ((newp (+ (position monster) (dp move-to))))
    (multiple-value-bind (newx newy) (point-to-coords newp)
      (when (point-to-coords (position monster))
        (setf (aref *monscreen* (x monster) (y monster)) nil))
      (if newx
          (if (occupied newx newy :excluding monster)
              (cancel-action move-to "Can't move here.")
              (progn
                (update-monscreen monster nil)
                (incf (position monster) (dp move-to))
                (cast (make-instance 'e-step :monster monster 
                                     :pt (position monster)))
                (update-monscreen monster t)))
          (incf (position monster) (dp move-to))))))

(defmethod do-action ((monster player) (move-to move-to))
  (with-slots (position) monster
    (with-slots (dp) move-to
      (when (> (abs dp) *max-moving-distance*)
        (return-from do-action nil))
      (multiple-value-bind (newx newy) (point-to-coords (+ position dp))
        (unless newx
          (do-action monster (make-instance 'center-screen))
          (setf (values newx newy) (point-to-coords (+ position dp))))
        (let ((monster? (aref *monscreen* newx newy)))
          (if (and monster? (not (eql monster? monster)))
              (cancel-action move-to "Can't move here.")
              ;;(instead-do-action move-to monster 
              ;;                   (make-instance 'attack :monster monster?))
              (call-next-method)))))))

(defmethod do-action :after ((monster player) (move-to move-to))
  (unless *cancelled*
    (update-fov)))


(defclass move-to-coords (action)
  ((dx :initarg :dx :accessor dx)
   (dy :initarg :dy :accessor dy)))

(defmethod do-action ((monster player) (move move-to-coords))
  (instead-do-action move monster
                     (make-instance 'move-to
                                    :dp (- 
                                         (coords-to-point (+ (x monster) (dx move))
                                                          (+ (y monster) (dy move)))
                                         (position monster)))))

(defclass move-to-coords-turning (action)
  ((dx :initarg :dx :accessor dx)
   (dy :initarg :dy :accessor dy)))

(defmethod do-action ((monster player) (move move-to-coords-turning))
  (let ((dp (- (coords-to-point (+ (x monster) (dx move))
                                (+ (y monster) (dy move)))
               (position monster))))
    (do-action monster (make-instance 'turn :angle (/ dp (direction monster))))
    (instead-do-action move monster (make-instance 'move-to :dp dp))))

(defclass move-forward (action) ())

(defmethod do-action ((monster monster) (move-forward move-forward))
  (instead-do-action move-forward monster
                     (make-instance 'move-to 
                                    :dp (* *moving-distance*
                                           (signum (direction monster))))))

(defclass step-back (action) ())

(defmethod do-action ((monster monster) (step-back step-back))
  (instead-do-action step-back monster
                     (make-instance 'move-to 
                                    :dp (* (- *moving-distance*)
                                           (signum (direction monster))))))

(defclass turn (action) 
  ((angle :initarg :angle :accessor angle) 
   (time :initform 1)))

(defmethod do-action ((monster monster) (turn turn))
  (setf (direction monster) (* (direction monster) (signum (angle turn)))))

(defmethod do-action :after ((monster player) (turn turn))
  (update-fov))

(defclass move-left (action) ())

(defmethod do-action ((monster monster) (move move-left))
  (do-action monster (make-instance 'turn :angle #C(0.0 1.0)))
  (instead-do-action move monster
                     (make-instance 'move-forward)))


(defclass move-right (action) ())

(defmethod do-action ((monster monster) (move move-right))
  (do-action monster (make-instance 'turn :angle #C(0.0 -1.0)))
  (instead-do-action move monster
                     (make-instance 'move-forward)))

(defclass strafe-left (action) ())

(defmethod do-action ((monster monster) (move strafe-left))
  (instead-do-action move monster
                     (make-instance 'move-to
                                    :dp (* *moving-distance*
                                            #C(0.0 1.0)
                                            (signum (direction monster))))))

(defclass strafe-right (action) ())

(defmethod do-action ((monster monster) (move strafe-right))
  (instead-do-action move monster
                     (make-instance 'move-to
                                    :dp (* *moving-distance*
                                            #C(0.0 -1.0)
                                            (signum (direction monster))))))


(defclass wait (action) ((time :initform 5)))

(defmethod do-action (monster (action wait)) 
  (declare (ignore action))
  (when (eql monster *player*)
    (make-instance 'e-wait :monster monster :pt (position monster))) 
  nil)

(defclass center-screen (action) ((time :initform 0)))

(defmethod do-action ((player player) (action center-screen))
  (with-slots (vector1 vector2) *screen-base*
    (change-base (- (+ (position player)
                       (/ vector1 *width* 2)
                       (/ vector2 *height* 2))
                    (/ vector1 2) (/ vector2 2))
                 vector1 vector2)))

(defclass turn-screen (action) 
  ((time :initform 0) 
   (angle :initarg :angle :initform 1.0 :accessor angle)))

(defmethod do-action ((player player) (action turn-screen))
  (with-slots ((v1 vector1) (v2 vector2)) *screen-base*
    (let* ((turn (signum (* (/ (direction player) v2) (angle action))))
           (vector1 (* v1 turn))
           (vector2 (* v2 turn)))
      (change-base (- (+ (position player)
                         (/ vector1 *width* 2)
                         (/ vector2 *height* 2))
                      (/ vector1 2) (/ vector2 2))
                   vector1 vector2))))

(defclass zoom (action) 
  ((time :initform 0)
   (factor :initarg :factor :accessor factor)))

(defmethod do-action ((player player) (zoom zoom))
  (with-slots ((v1 vector1) (v2 vector2)) *screen-base*
    (let* ((vector1 (* v1 (factor zoom)))
           (vector2 (* v2 (factor zoom))))
      (when (or (> (abs vector1) *maxzoom*)
                (< (abs vector1) *minzoom*))
        (cancel-action zoom "You DON'T need to zoom any further.")
        (return-from do-action))
      (decf *zoom-level* (round (signum (1- (factor zoom)))))
      (change-base (- (+ (position player)
                         (/ vector1 *width* 2)
                         (/ vector2 *height* 2))
                      (/ vector1 2) (/ vector2 2))
                   vector1 vector2))))
                        
(defclass quit-game (action) ((time :initform 0)))

(defmethod do-action ((monster player) (quit quit-game))
  (if (yes-or-no "Do you really want to quit?")
      (setf *quitflag* :quit)
      (cancel-action quit "OK, let's continue...")))

(defclass no-action (action)
  ((key :initarg :key :initform 0 :accessor key)
   (time :initform 0)))

(defmethod do-action ((monster player) (action no-action))
  (setf *zero-time-action* t)
  (message "No action bound to key: ~a" (key action)))

(defclass cancel (wait) ((time :initform 0)))

(defclass next-message (action) ((time :initform 0)))

(defmethod do-action ((monster player) (action next-message))
  (setf *zero-time-action* t)
  ;(format t "~a~%" *messages*)
  (display-message (pop *messages*)))

(defclass recap (action) ((time :initform 0)))

(defmethod do-action ((monster player) (action recap))
  (recap-messages)
  (getch)
  (redraw-screen))

(defclass toggle-fov (action) ((time :initform 0)))

(defmethod do-action ((monster player) (action toggle-fov))
  (if *show-fov*
      (progn (setf *show-fov* nil)
             (redraw-screen))
      (setf *show-fov* t)))

(defclass keychord-tool (action) ((time :initform 0)))

(defmethod do-action ((monster player) (action keychord-tool))
  (display-simple-message "Press the key, or several keys, followed by Enter")
  (loop with keychord
       for x = (getch)
       until (= x 13)
       do (push x keychord)
       finally 
       (message 
        (format nil "Your keychord is: ~a"
                (if (= (length keychord) 1) 
                    (car keychord) (nreverse keychord))))))

(defclass cheat-action (action) ((time :initform 0)))

(defmethod do-action :around ((monster player) (action cheat-action))
  (when *debug-mode*
    (call-next-method)))

(defclass restore-hp (cheat-action) ())

(defmethod do-action ((monster player) (action restore-hp))
  (setf (hp monster) (max-hp monster)))

(defclass show-repl-console (cheat-action) ())

(defmethod do-action ((monster player) (action show-repl-console))
  (repl-console 15)
  (redraw-screen)) 

;;SECTION 7a: Combat

(defclass attack (action)
  ((monster :initarg :monster :accessor monster)
   (weapon :initarg :weapon :accessor weapon)
   (time :initform 6)))

(defun hit-chance (angle distance)
  (* (sqrt (cos angle)) (/ 0.5 (+ 0.5 distance))))

(defun hit-p (weapon distance angle)
  (case weapon
    (:rifle
     (< (random 1.0) (hit-chance angle distance)))
    (:knife t)))

(defun calculate-hit (monster target)
  (let* ((p1 (position monster))
         (p2 (position target))
         (dist (distance p1 p2))
         (angle (abs (phase (/ (- p2 p1) (direction monster))))))
    (hit-chance angle dist)))

(defun weapon-damage (weapon)
  (case weapon
    (:rifle (1+ (random 3)))
    (:knife 5)))

(defmethod do-action (monster (attack attack))
  (with-slots ((monster2 monster)) attack
    (let* ((p1 (position monster))
           (p2 (position monster2))
           (dist (distance p1 p2))
           (angle (abs (phase (/ (- p2 p1) (direction monster))))))
      (if (hit-p (weapon attack) dist angle)
          (let ((dam (weapon-damage (weapon attack))))
            (decf (hp monster2) dam)
            (cast (make-instance 'e-c-hit 
                                 :attacker monster
                                 :defender monster2
                                 :damage dam
                                 :weapon (weapon attack))))
          (cast (make-instance 'e-c-miss
                               :attacker monster
                               :defender monster2
                               :weapon (weapon attack))))
      (when (<= (hp monster2) 0) (kill monster2)))))
          

(defclass shoot (action) ((time :initform 6)))

(defmethod do-action ((monster player) (action shoot))
  (let ((targets (visible-monsters)))
    (let ((target (choose-target targets (mapcar (lambda (tgt)
                                                   (calculate-hit *player*
                                                                  (cdr tgt)))
                                                 targets))))
      (if target
          (instead-do-action action monster
                             (make-instance 'attack :monster target
                                            :weapon :rifle))
          (cancel-action action "There's no one to shoot at or no target chosen.")))))
            
(defclass stab (action) ((time :initform 6)))

(defmethod do-action ((monster player) (action stab))
  (let ((targets (loop for x from 0 below *width* nconcing
                      (loop for y from 0 below *height*
                         for mon = (aref *monscreen* x y)
                         for fov = (bit *fovmap* x y)
                         when (and mon (not (eql mon *player*)) 
                                   (= fov 1) (< (distance (position mon)
                                                          (position *player*))
                                                *stab-range*))
                         collect (cons (cons x y) mon)))))
    (let ((target (choose-target targets)))
      (if target
          (instead-do-action action monster
                             (make-instance 'attack :monster target
                                            :weapon :knife))
          (cancel-action action "There's no one close enough or no target chosen.")))))

;;SECTION 8: Events

(defun aware-p (monster pt &aux (dist (distance (position monster) pt)))
  (when (eql monster *player*)
    (return-from aware-p (or (< dist *intuition-range*)
                             (orient *curfov* pt))))
  (and (< dist *sight-range*)
       (< (abs (phase (/ (- pt (position monster)) (direction monster))))
          *fov-angle-phase*)
       (let ((line (make-line (position monster) pt))
             (secs (neighbors (point-to-sector (position monster)))))
         (loop for sec in secs
              always (loop for obj in (gethash sec *map-sector-hash*)
                        never (and (housep obj)
                                   (intersect-p line (area obj))))))))
 
(defclass event () 
  ((interesting :initform t :initarg :interesting :accessor interesting)
   (pt :initarg :pt :accessor pt)))

(defgeneric cast (event)
  (:documentation "Make all monsters that feel the event react to it."))

(defmethod cast :before (event)
  (declare (ignore event))
  (push *curmonster*  *monsters*))

(defmethod cast ((event event)) 
  (when (aware-p *player* (pt event))
    (react *player* event)))
      
;;  (loop for mon in *monsters*
;;        when (aware-p mon (pt event))
;;        do (react mon event)))

(defmethod cast :after (event)
  (declare (ignore event))
  (pop *monsters*))

(defclass e-step (event) 
  ((interesting :initform nil)
   (monster :initarg :monster :accessor monster)))

(defmethod cast ((event e-step))
  (loop for mon in *monsters*
       when (< (distance (position mon) (pt event)) *hearing-range*)
       do (react mon event)))

(defclass e-wait (event) 
  ((interesting :initform nil)
   (monster :initarg :monster :accessor monster)))



(defmethod cast ((event e-shout))
  ;;(format t "Casting shout: ~%")
  (loop for mon in *monsters*
     when (< (distance (position mon) (pt event)) *shouting-range*)
     do ;;(format t "~a ~%" mon)
       (react mon event)))

(defmethod description ((event e-shout))
  (format nil "~@(~a~) shouts: \"I see him!\"." (name (monster event))))

(defclass e-combat (event)
  ((attacker :initarg :attacker :accessor attacker)
   (defender :initarg :defender :accessor defender)
   (weapon :initarg :weapon :accessor weapon)))

(defmethod cast ((event e-combat))
  (case (weapon event)
    (:knife (react *player* event))
    (:rifle 
     (loop for mon in *monsters*
          when (or (< (distance (position mon) (position (attacker event)))
                      *shouting-range*)
                   (< (distance (position mon) (position (defender event)))
                      *shouting-range*))
          do (react mon event)))))

(defgeneric combat-situation (attacker defender weapon result)
  (:documentation "Describe a combat situation"))

(defgeneric attack-word (weapon form)
  (:documentation "Returns a verb that would be used in a combat message")
  (:method (weapon form) (declare (ignore weapon))
           (if (eql form :he) "hits" "hit")))

(defmethod attack-word ((weapon (eql nil)) form)
  (if (eql form :he) "punches" "punch"))

(defmethod attack-word ((weapon (eql :knife)) form)
  (if (eql form :he) "stabs" "stab"))

(defmethod attack-word ((weapon (eql :rifle)) form)
  (if (eql form :he) "shoots at" "shoot at"))

(defmethod combat-situation (att def weapon result)
  (format nil "~@(~a~) ~a ~a~a." 
                 (name att) 
                 (if (eql result :hit) 
                     (attack-word weapon :he)
                     (format nil "tries to ~a" 
                             (attack-word weapon :inf)))
                 (name def)
                 (if (eql result :hit) "" ", but misses")))

(defmethod combat-situation ((att player) def weapon result)
  (format nil "~@(~a~) ~a ~a~a." 
                 (name att) 
                 (if (eql result :hit) 
                     (attack-word weapon :you)
                     (format nil "try to ~a" 
                             (attack-word weapon :inf)))
                 (name def)
                 (if (eql result :hit) "" ", but miss")))

(defmethod combat-situation (att (def player) weapon result)
  (format nil "~@(~a~) ~a ~a~a." 
                 (name att) 
                 (if (eql result :hit) 
                     (attack-word weapon :he)
                     (format nil "tries to ~a" 
                             (attack-word weapon :inf)))
                 "you"
                 (if (eql result :hit) "" ", but misses")))


(defclass e-c-hit (e-combat)
  ((damage :initarg :damage :accessor damage)))

(defmethod description ((event e-c-hit))
  (with-slots ((att attacker) (def defender) weapon) event
    (combat-situation att def weapon :hit)))

(defclass e-c-miss (e-combat) ())

(defmethod description ((event e-c-miss))
  (with-slots ((att attacker) (def defender) weapon) event
    (combat-situation att def weapon :miss)))

(defclass e-monster (event)
  ((monster :initarg :monster :accessor monster)))

(defmethod cast ((event e-monster))
  (when (aware-p *player* (position (monster event)))
    (react *player* event)))

;;  (loop for mon in *monsters*
;;       when (aware-p mon (position (monster event)))
;;        do (react mon event)))

(defclass e-dead (e-monster) ())

(defmethod description ((event e-dead))
  (if (eql (monster event) *player*)
      (progn
        (show-status)
        (display-simple-message "You die..."))
      (let ((dead-message 
             (list "~@(~a~) dies.")))
        (format nil (nth (random (length dead-message)) dead-message) 
                (name (monster event))))))

(defclass e-m-angry (e-monster) ())

(defmethod description ((event e-m-angry))
  (format nil "~@(~a~) notices you." (name (monster event))))

;;SECTION 9: Monster loop processing

(defun add-to-stack (monster)
  (if (and *zero-time-action* (eql monster *player*))
      (push monster *monsters*)
      (setf *monsters* (nconc *monsters* (list monster)))))

(defun run-ai (monster)
  (if (> (distance (position monster) (position *player*)) *monster-range*)
      (progn (setf *zero-time-action* nil) (add-to-stack monster))
      (progn
        (let ((action (ai monster)))
          (do-action monster action)
          ;;(when (and (= (random 10) 0)
          ;;           (< (hp monster) (max-hp monster)))
          ;;  (incf (hp monster)))
          )
        (add-to-stack monster))))

(defun run-stack ()
  (unless *monsters* (return-from run-stack :quit))
  (run-ai (setf *curmonster* (pop *monsters*)))
  (return-from run-stack :ok))

;;SECTION 10: Main sequence

(defun splash-screen ()
  (erase)
  (attrset :crose)
  (mvprintw 1 1 "Urban Warfare: The Escape 0.8")
  (attrset :cwhite)
  (mvprintw 2 1 "(c) 2006 Timofei Shatrov")
  (attrset :cyellow)
  (mvprintw 3 1 "2007 Annual 7DRL Challenge")
  (attrset :cwhite)
  (mvprintw 6 1 "You are (or were) a pilot, returning to home base after blowing some stuff up.")
  (mvprintw 7 1 "Unfortunately you were shot down...")
  (mvprintw 8 10 "...and you landed smack down in the middle of the enemy city!")
  (mvprintw 9 1 "You have to use your wits to get outta here!")
  (mvprintw 10 1 "You also got a rifle and a knife.")
  (attrset :cyellow)
  (mvprintw 12 1 "Controls")
  (attrset :cgray)
  (mvprintw 14 1 "Q (turn left)   W (forward)    E (turn right)    R (shoot rifle) T (turn scr)")
  (mvprintw 15 1 "A (strafe left) S (step back)  D (strafe right)  F (show enemy FOV)")
  (mvprintw 16 1 "Z (move left)   X (center scr) C (move right)              B (stab stab stab)")
  
  (mvprintw 18 1 "Numpad (move coordinate-wise) Ctrl+Numpad or 0+Numpad (turn+move)")
  (mvprintw 19 1 "+/- (zoom scr in/out)      5/. (wait)  Shift+Q (quit game)")
  (mvprintw 20 1 "Space (next message)       Backspace (recap messages)")
  (attrset :crose)
  (mvprintw 22 1 "Every key is important!")
  (attrset :cwhite)
  (mvprintw 24 50 "[press any key to continue]")
  (refresh)
  (getch))


(defun genmd5 (object)
  (coerce (md5:md5sum-sequence (let ((*print-readably* t)) (format nil "~s" object)))
          'list))

(defun load-hiscore ()
  (unless (probe-file "hiscore") (return-from load-hiscore nil))
  (with-open-file (in "hiscore"
                      :external-format #+clisp charset:utf-8 #+sbcl :utf-8)
    (let* ((*read-eval* nil)
           (hashentry (read in))
           (hiscore (read in)))
      (if (equal (genmd5 hiscore) hashentry) hiscore :error))))

(defun save-hiscore (hiscore)      
  (with-open-file (out "hiscore" :direction :output :if-exists :supersede
                       :external-format #+clisp charset:utf-8 #+sbcl :utf-8)
    (let ((*print-readably* t))
      (format out "~s~%~s " (genmd5 hiscore) hiscore))))

(defun compare-scores (a b)
  "A>B if (CAR A)>(CAR B) or (CAR A)=(CAR B) and (CDR A)<(CDR B)"
  (or (> (car a) (car b))
      (and (= (car a) (car b)) (< (cdr a) (cdr b)))))

(defun hiscore (&aux hiscore (number 0))
  (setf hiscore (load-hiscore))
  (when (eql hiscore :error)
    (erase)
    (attrset :cred)
    (mvprintw 1 1 "Your hiscore file is corrupted. Please delete it.")
    (attrset :cwhite)
    (mvprintw 24 50 "[press any key to quit]")
    (getch)
    (return-from hiscore))
    
  (when (eql *quitflag* :escape)
    (setf hiscore (nreverse hiscore))
    (setf hiscore (stable-sort (nreverse (push (cons :player 
                                                     (cons *kill-count*
                                                           *turn-count*))
                                               hiscore))
                               #'compare-scores :key #'cdr)))
  (when (> (length hiscore) 10)
    (setf hiscore (nbutlast hiscore)))
  (let ((ask (loop for entry in hiscore
                   do (incf number)
                   when (eql (car entry) :player) do (loop-finish)
                   finally (if (eql (car entry) :player) (return entry) nil))))
             ;(find :player hiscore :key #'car)))
    (when ask
      (erase) (attrset :cyellow)
      (mvprintw 1 1 "Congratulations! You made the top ten!")
      (attrset :grey) (mvprintw 2 1 "Enter your name: ")
      (setf (car ask) (getnstr 12)))
    (unless ask (setf number 0)))

  (erase)
  (attrset :crose)
  (mvprintw 0 1 "Top 10 Players:")
  (attrset :cgray)
  (mvprintw 1 1 (format nil "NN. ~15a~10a ~a" "Name" "Kills" "Turns"))
  (mvprintw 2 1 (format nil "~,,40,'-A" ""))
  (attrset :cwhite)
  (loop for i from 1 to 10
        for entry in hiscore
        when (= i number) do (attrset :cyellow)
        do (mvprintw (+ 2 i) 1 (format nil "~2d. ~15a~10a ~a" 
                                       i (car entry) (cadr entry)
                                       (cddr entry)))
        when (= i number) do (attrset :cwhite))
  (save-hiscore hiscore)

  (mvprintw 24 50 "[press any key to quit]")
  (refresh)
  (getch))


(defun runtime ()
  (setf *quitflag* nil)
  (loop 
     (run-stack) 
     (when *quitflag*
       (unless (eql *quitflag* :quit) (hiscore))
       ;;(morgue-dump)
       (unless (eql *quitflag* :quit) (getch))
       (return))))

(defun start-game (&key debug)
  (setf *random-state* (make-random-state t)
        *allmessages* nil
        *messages* nil
        *debug-mode* debug
        *curmonster* nil
        *monsters* nil
        ;;*monster-sector-hash* (make-hash-table :test #'equal)
        *player* nil
        *show-fov* nil
        *kill-count* 0
        *turn-count* 0
        *zoom-level* 0)
  (connect-console)
  (erase)
  (splash-screen)
  (unwind-protect
       (progn
         (when (init-controls) (error "Cannot find control file. This sucks :("))
         (erase)
         (setf *screen-base* (make-base #C(0.0 0.0) #C(0.6 0.0) #C(0.0 0.3)))
         (generate-street-map 500)
         (process-streets)
         (change-base #C(0.0 0.0) #C(0.75 0.0) #C(0.0 0.375))
         (generate-player)
         (redraw-screen)
         (refresh-screen)
         (runtime))
    (unless *debug-mode* (close-console)))
  (unless debug 
    #+clisp (ext:quit)
    #+sbcl (sb-ext:quit)))

#+clisp
(defun make-exec ()
  (ext:saveinitmem "uwar.exe" 
                   :quiet t :norc t :init-function #'start-game 
                   :start-package :uwar :executable t))
